# TODO: use final (https://docs.python.org/3/library/typing.html#typing.Final) after migrating to python 3.8 or newer

PEARSON_NAME = "Pearson Correlation"
KENDALLTAU_NAME = "Kendall\'s Tau Correlation"
SPEARMAN_NAME = "Spearman\'s R Correlation"
POINTBISERIAL_NAME = "Pointbiserial Correlation"
STUDENTS_T_NAME = "Student\'s T Test"
PAIRED_STUDENTS_NAME = "Paired Student\'s T Test"
WELCHS_T_NAME = "Welch\'s T Test"
MANN_WHITNEY_NAME = "Mann Whitney U Test"
WILCOXON_SIGNED_RANK_NAME = "Wilcoxon Signed Rank Test"
RM_ONE_WAY_ANOVA_NAME = "Repeated Measures One Way ANOVA"
FACTORIAL_ANOVA_NAME = "Factorial ANOVA"
KRUSKALL_WALLIS_NAME = "Kruskall Wallis"
F_TEST_NAME = "F Test"
CHI_SQUARE_NAME = "Chi Square Test"
FISHER_EXACT_NAME = "Fisher\'s Exact Test"
