from tea.api import (
                    data, 
                    define_variables, 
                    define_study_design, 
                    assume,
                    hypothesize,
                    download_data,
                    divine_properties
                )